#ifndef __SIMPLE_MESH__
#define __SIMPLE_MESH__

#include <string>
#include <vector>
#include <float.h>
#include <iostream>
#include <fstream>

#include "geom.h"
#include <glm/glm.h>

// #include "GLGPUmesh.h" // must come after QT

class Simple_mesh{

public:
  typedef float        real_type;
  typedef unsigned int size_type;
  typedef unsigned int index_type;

  typedef Geometry::Vec3<real_type>  real3_type;
  typedef Geometry::Vec3<index_type> index3_type;
  
  typedef std::vector<index3_type> index3_list_type;
  typedef std::vector<real3_type>  real3_list_type;
  typedef std::pair<Geometry::Vec3f,Geometry::Vec3f> Bbox;

  inline Simple_mesh();

  inline void load(const std::string& file_name, bool unitize = false);
  inline void save( const char *fileName );

  inline void update();
  
  inline void draw() const;

  inline size_type vertexSize() const;
  inline size_type normalSize() const;
  inline size_type faceSize() const;
  
  real3_list_type  vertex;
  real3_list_type  vertex_normal;
  index3_list_type triangle_vertex_index;
  index3_list_type triangle_normal_index;
  Bbox bbox;

private:
//   typedef GPUMESH_MVF2(GPUmesh::mvf_vertex_3f,
// 		       GPUmesh::mvf_normal_3f) GPUmesh_vertex_type;
  
//   typedef GPUmesh::GPUmesh_GL_DisplayList<GPUmesh_vertex_type>
//   GPUmesh_display_type;

// // typedef GPUMESH_MVF2(GPUmesh::mvf_vertex_2f,GPUmesh::mvf_color0_rgba) GPUmesh_vertex_type;
// //   typedef GPUmesh::GPUmesh_GL_DisplayList<GPUmesh_vertex_type>
// //   GPUmesh_display_type;
  
//   GPUmesh_display_type* display;

};









/*
  
  #############################################
  #############################################
  #############################################
  ######                                 ######
  ######   I M P L E M E N T A T I O N   ######
  ######                                 ######
  #############################################
  #############################################
  #############################################
  
*/





Simple_mesh::Simple_mesh()
//   display(NULL)
{}


void Simple_mesh::load(const std::string& file_name, bool unitize/* = false*/){

  GLMmodel *model;
  GLfloat *modelNormals;
  int numNormals;
  bool useFacet = false;
  model = glmReadOBJ(file_name.c_str());

  if (unitize) {
	glmUnitize(model);
  }

  // If there are no normals
  if (model->numnormals == 0) {
	  // generate facet normals at each vertex
	  glmFacetNormals(model);
	  numNormals = model->numfacetnorms;
	  modelNormals = model->facetnorms;
	  useFacet = true;
	  //glmVertexNormals(model, 90);
	  //numNormals = model->numnormals;
	  //modelNormals = model->normals;
  }
  else {
	  numNormals = model->numnormals;
	  modelNormals = model->normals;
  }
  vertex.resize(model->numvertices);
  vertex_normal.resize(numNormals);

  // Initialize the bounding box for the mesh
  for (int d = 0; d < 3; d++) {
	  bbox.first[d] = FLT_MAX;
	  bbox.second[d] = -FLT_MAX;
  }		

  // Assign vertices
  // GLM reads into a 1-based array
  for(size_type v = 0,v_end = vertex.size();v < v_end;v++){
    for(size_type n = 0;n < 3;n++){
      vertex[v][n]        = model->vertices[3 * (v+1) + n];

	  // Calculate the bounding box of the mesh
	  if (vertex[v][n] < bbox.first[n]) {
		  bbox.first[n] = vertex[v][n];
	  }
	  else if (vertex[v][n] > bbox.second[n]) {
		  bbox.second[n] = vertex[v][n];
	  }
	}
  }

  // Assign normals
  // GLM reads into a 1-based array
  for(size_type v = 0,v_end = vertex_normal.size();v < v_end;v++){
	  for(size_type n = 0;n < 3;n++){
		vertex_normal[v][n] = modelNormals[3 * (v+1) + n];
	  }
  }

  const size_type max_v = vertex.size()-1;
  size_type index = 0;
  
  triangle_vertex_index.resize(model->numtriangles);
  triangle_normal_index.resize(model->numtriangles);

  // GLM reads into a 1-based array
  for(size_type t = 0,t_end = triangle_vertex_index.size();
      t < t_end;++t){

    GLMtriangle& triangle = model->triangles[t];

    index3_type& tvi = triangle_vertex_index[index];
	index3_type& tni = triangle_normal_index[index];
    
    bool valid = true;
    
    for(size_type i = 0;i < 3;i++){
      const index_type tvii = triangle.vindices[i]-1;
      tvi[i] = tvii;
	  tni[i] = (useFacet ? t : triangle.nindices[i]-1);

      valid = valid && (tvii <= max_v);
    }

    if (valid) index++;
  } // END OF for t
  
  triangle_vertex_index.resize(index);
  triangle_normal_index.resize(index);
  
  glmDelete(model);

  update();
}




void Simple_mesh::update() {
/* 
  if (display != NULL) delete display;

  display = new GPUmesh_display_type();
  
  display->begin(GPUMESH_TRIANGLELIST);
  
  for(size_type t = 0,t_end = triangle_vertex_index.size();t < t_end;t++){

    const index3_type& tvi = triangle_vertex_index[t];
    const index3_type& tni = triangle_normal_index[t];

    for(size_type i = 0;i < 3;i++){

      const real3_type& v = vertex[tvi[i]];
      const real3_type& n = vertex_normal[tni[i]];
    
      display->normal(n[0],n[1],n[2]);
      display->vertex_3(v[0],v[1],v[2]);
    }
  }
//   display->color0_4(0,0,255,0);
//   display->vertex_2(0.1,0.1);

//   display->color0_4(255,0,0,0);
//   display->vertex_2(0.1,0.9);

//   display->color0_4(0,255,0,0);
//   display->vertex_2(0.9,0.9);

//   display->color0_4(0,0,255,0);
//   display->vertex_2(0.1,0.1);

//   display->color0_4(0,255,0,0);
//   display->vertex_2(0.9,0.9);

//   display->color0_4(64,0,255,0);
//   display->vertex_2(0.9,0.1);  
  display->end();
*/
}

void Simple_mesh::draw() const{

//   display->render();
  
  glBegin(GL_TRIANGLES);

  for(size_type t = 0,t_end = triangle_vertex_index.size();t < t_end;t++){

    const index3_type& tvi = triangle_vertex_index[t];
    const index3_type& tni = triangle_normal_index[t];

    for(size_type i = 0;i < 3;i++){

      const real3_type& v = vertex[tvi[i]];
      const real3_type& n = vertex_normal[tni[i]];
    
      glNormal3f(n[0],n[1],n[2]);
      glVertex3f(v[0],v[1],v[2]);
    }
  }
  
  glEnd();
}


inline Simple_mesh::size_type Simple_mesh::vertexSize() const {
	return vertex.size();
}

inline Simple_mesh::size_type Simple_mesh::normalSize() const {
	return vertex_normal.size();
}

inline Simple_mesh::size_type Simple_mesh::faceSize() const {
	return triangle_vertex_index.size();
}

void Simple_mesh::save( const char *fileName )
{
	// Based on http://www.nada.kth.se/~asa/Ray/vertface2obj.m

	std::ofstream out;
	out.open(fileName);

	if (out.fail()) {
		std::cerr << "Error in Simple_mesh::save(): Could not open output file " << fileName << " for writing!" << std::endl;
		return;
	}

	std::cout << "Saving vertex and triangle data in " << fileName << " ...";

	// Write vertex data
	for (index_type i = 0; i < vertex.size(); i++) {
		out << "v";
		for (int c = 0; c < 3; c++) {
			out << " " << vertex[i][c];
		}
		out << "\n";
	}

	// Write triangle data
	out << "g foo\n";
	for (index_type i = 0; i < triangle_vertex_index.size(); i++) {
		out << "f";
		for (int c = 0; c < 3; c++) {
			out << " " << triangle_vertex_index[i][c]+1;
		}
		out << "\n";
	}
	out << "g\n";

	out.close();

	std::cout << "...done!" << std::endl;
}

#endif
